@@ -1,4 +1,4 @@
-/* $Id: dnsget.c,v 1.18 2005/05/16 12:43:26 mjt Exp $
+/* dnsget.c
    simple host/dig-like application using UDNS library
 
    Copyright (C) 2005  Michael Tokarev <mjt@corpit.ru>
@@ -21,20 +21,35 @@
 
  */
 
-#include <stdio.h>
-#include <unistd.h>
-#include <stdlib.h>
-#include <string.h>
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+#ifdef WINDOWS
+#include <windows.h>
+#include <winsock2.h>
+#else
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
-#include <arpa/inet.h>
-#include <time.h>
 #include <sys/time.h>
+#include <unistd.h>
+#endif
+#include <time.h>
 #include <stdarg.h>
 #include <errno.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include "udns.h"
 
+#ifndef HAVE_GETOPT
+# include "getopt.c"
+#endif
+
+#ifndef AF_INET6
+# define AF_INET6 10
+#endif
+
 static char *progname;
 static int verbose = 1;
 static int errors;
@@ -58,6 +73,11 @@ static void die(int errnum, const char *fmt, ...) {
   exit(1);
 }
 
+static const char *dns_xntop(int af, const void *src) {
+  static char buf[6*5+4*4];
+  return dns_ntop(af, src, buf, sizeof(buf));
+}
+
 struct query {
   const char *name;		/* original query string */
   unsigned char *dn;		/* the DN being looked up */
@@ -101,7 +121,7 @@ printtxt(const unsigned char *c) {
   unsigned n = *c++;
   const unsigned char *e = c + n;
   if (verbose > 0) while(c < e) {
-    if (*c < ' ' || *c >= 127) printf("\\%02x", *c);
+    if (*c < ' ' || *c >= 127) printf("\\%03u", *c);
     else if (*c == '\\' || *c == '"') printf("\\%c", *c);
     else putchar(*c);
     ++c;
@@ -167,8 +187,12 @@ printrr(const struct dns_parse *p, struct dns_rr *rr) {
   if (verbose > 0) {
     if (verbose > 1) {
       if (!p->dnsp_rrl && !rr->dnsrr_dn[0] && rr->dnsrr_typ == DNS_T_OPT) {
-        printf(";EDNS0 OPT record (UDPsize: %d): %d bytes\n",
-               rr->dnsrr_cls, rr->dnsrr_dsz);
+        printf(";EDNS%d OPT record (UDPsize: %d, ERcode: %d, Flags: 0x%02x): %d bytes\n",
+               (rr->dnsrr_ttl>>16) & 0xff,	/* version */
+               rr->dnsrr_cls,			/* udp size */
+               (rr->dnsrr_ttl>>24) & 0xff,	/* extended rcode */
+               rr->dnsrr_ttl & 0xffff,		/* flags */
+               rr->dnsrr_dsz);
         return;
       }
       n = printf("%s.", dns_dntosp(rr->dnsrr_dn));
@@ -203,7 +227,7 @@ printrr(const struct dns_parse *p, struct dns_rr *rr) {
 
   case DNS_T_AAAA:
     if (rr->dnsrr_dsz != 16) goto xperr;
-    printf("%s", inet_ntop(AF_INET6, dptr, dn, DNS_MAXDN));
+    printf("%s", dns_xntop(AF_INET6, dptr));
     break;
 
   case DNS_T_MX:
@@ -241,7 +265,7 @@ printrr(const struct dns_parse *p, struct dns_rr *rr) {
   case DNS_T_WKS:
     c = dptr;
     if (dptr + 4 + 2 >= end) goto xperr;
-    printf("%s %d", inet_ntoa(*((struct in_addr*)dptr)), dptr[4]);
+    printf("%s %d", dns_xntop(AF_INET, dptr), dptr[4]);
     c = dptr + 5;
     for (n = 0; c < dend; ++c, n += 8) {
       if (*c) {
@@ -262,7 +286,31 @@ printrr(const struct dns_parse *p, struct dns_rr *rr) {
            dns_dntosp(dn));
     break;
 
-  case DNS_T_KEY: /* flags(2) proto(1) algo(1) pubkey */
+  case DNS_T_NAPTR:	/* order pref flags serv regexp repl */
+    c = dptr;
+    c += 4;	/* order, pref */
+    for (n = 0; n < 3; ++n)
+      if (c >= dend) goto xperr;
+      else c += *c + 1;
+    if (dns_getdn(pkt, &c, end, dn, DNS_MAXDN) <= 0 || c != dend) goto xperr;
+    c = dptr;
+    printf("%u %u", dns_get16(c+0), dns_get16(c+2));
+    c += 4;
+    for(n = 0; n < 3; ++n) {
+      putchar(' ');
+      if (verbose > 0) putchar('"');
+      c = printtxt(c);
+      if (verbose > 0) putchar('"');
+    } 
+    printf(" %s.", dns_dntosp(dn));
+    break;
+
+  case DNS_T_KEY:
+  case DNS_T_DNSKEY:
+    /* flags(2) proto(1) algo(1) pubkey */
+  case DNS_T_DS:
+  case DNS_T_DLV:
+    /* ktag(2) proto(1) algo(1) pubkey */
     c = dptr;
     if (c + 2 + 1 + 1 > dend) goto xperr;
     printf("%d %d %d", dns_get16(c), c[2], c[3]);
@@ -274,12 +322,13 @@ printrr(const struct dns_parse *p, struct dns_rr *rr) {
     break;
 
   case DNS_T_SIG:
+  case DNS_T_RRSIG:
     /* type(2) algo(1) labels(1) ottl(4) sexp(4) sinc(4) tag(2) sdn sig */
     c = dptr;
     c += 2 + 1 + 1 + 4 + 4 + 4 + 2;
     if (dns_getdn(pkt, &c, end, dn, DNS_MAXDN) <= 0) goto xperr;
-    printf("%d %u %u %u ",
-           dns_get16(dptr), dptr[2], dptr[3], dns_get32(dptr+4));
+    printf("%s %u %u %u ",
+           dns_typename(dns_get16(dptr)), dptr[2], dptr[3], dns_get32(dptr+4));
     printdate(dns_get32(dptr+8));
     putchar(' ');
     printdate(dns_get32(dptr+12));
@@ -287,15 +336,14 @@ printrr(const struct dns_parse *p, struct dns_rr *rr) {
     printb64(c, dend);
     break;
 
-#if 0	/* unused RR types? */
-  case DNS_T_DS:
-    c = dptr;
-    if (c + 2 + 2 >= dend) goto xperr;
-    printf("%u %u %u ", dns_get16(c), c[2], c[3]);
-    printhex(c + 4, dend);
+  case DNS_T_SSHFP: /* algo(1), fp type(1), fp... */
+    if (dend < dptr + 3) goto xperr;
+    printf("%u %u ", dptr[0], dptr[1]); /* algo, fp type */
+    printhex(dptr + 2, dend);
     break;
 
-  case DNS_T_NSEC:
+#if 0	/* unused RR types? */
+  case DNS_T_NSEC: /* nextDN bitmaps */
     c = dptr;
     if (dns_getdn(pkt, &c, end, dn, DNS_MAXDN) <= 0) goto xperr;
     printf("%s.", dns_dntosp(dn));
@@ -373,21 +421,15 @@ dbgcb(int code, const struct sockaddr *sa, unsigned slen,
   }
   else
     printf(";; received %d bytes response from ", r);
-  if (sa->sa_family == AF_INET && slen >= sizeof(struct sockaddr_in)) {
-    char buf[4*4];
+  if (sa->sa_family == AF_INET && slen >= sizeof(struct sockaddr_in))
     printf("%s port %d\n",
-           inet_ntop(AF_INET, &((struct sockaddr_in*)sa)->sin_addr,
-                     buf, sizeof(buf)),
+           dns_xntop(AF_INET, &((struct sockaddr_in*)sa)->sin_addr),
            htons(((struct sockaddr_in*)sa)->sin_port));
-  }
-#ifdef AF_INET6
-  else if (sa->sa_family == AF_INET6 && slen >= sizeof(struct sockaddr_in6)) {
-    char buf[6*5+4*4];
+#ifdef HAVE_IPv6
+  else if (sa->sa_family == AF_INET6 && slen >= sizeof(struct sockaddr_in6))
     printf("%s port %d\n",
-           inet_ntop(AF_INET6, &((struct sockaddr_in6*)sa)->sin6_addr,
-                     buf, sizeof(buf)),
+           dns_xntop(AF_INET6, &((struct sockaddr_in6*)sa)->sin6_addr),
            htons(((struct sockaddr_in6*)sa)->sin6_port));
-  }
 #endif
   else
     printf("<<unknown socket type %d>>\n", sa->sa_family);
@@ -417,10 +459,13 @@ dbgcb(int code, const struct sockaddr *sa, unsigned slen,
   printf(", status: %s, id: %d, size: %d\n;; flags:",
          dns_rcodename(dns_rcode(pkt)), dns_qid(pkt), r);
   if (dns_qr(pkt)) printf(" qr");
-  if (dns_rd(pkt)) printf(" rd");
-  if (dns_ra(pkt)) printf(" ra");
   if (dns_aa(pkt)) printf(" aa");
   if (dns_tc(pkt)) printf(" tc");
+  if (dns_rd(pkt)) printf(" rd");
+  if (dns_ra(pkt)) printf(" ra");
+  /* if (dns_z(pkt))  printf(" z"); only one reserved bit left */
+  if (dns_ad(pkt)) printf(" ad");
+  if (dns_cd(pkt)) printf(" cd");
   numqd = dns_numqd(pkt);
   printf("; QUERY: %d, ANSWER: %d, AUTHORITY: %d, ADDITIONAL: %d\n",
          numqd, dns_numan(pkt), dns_numns(pkt), dns_numar(pkt));
@@ -466,7 +511,7 @@ static void dnscb(struct dns_ctx *ctx, void *result, void *data) {
   struct dns_rr rr;
   unsigned nrr;
   unsigned char dn[DNS_MAXDN];
-  const unsigned char *pkt, *cur, *end, *qdn;
+  const unsigned char *pkt, *cur, *end;
   if (!result) {
     dnserror(q, r);
     return;
@@ -475,10 +520,9 @@ static void dnscb(struct dns_ctx *ctx, void *result, void *data) {
   dns_getdn(pkt, &cur, end, dn, sizeof(dn));
   dns_initparse(&p, NULL, pkt, cur, end);
   p.dnsp_qcls = p.dnsp_qtyp = 0;
-  qdn = dn;
   nrr = 0;
   while((r = dns_nextrr(&p, &rr)) > 0) {
-    if (!dns_dnequal(qdn, rr.dnsrr_dn)) continue;
+    if (!dns_dnequal(dn, rr.dnsrr_dn)) continue;
     if ((qcls == DNS_C_ANY || qcls == rr.dnsrr_cls) &&
         (q->qtyp == DNS_T_ANY || q->qtyp == rr.dnsrr_typ))
       ++nrr;
@@ -491,10 +535,10 @@ static void dnscb(struct dns_ctx *ctx, void *result, void *data) {
       }
       else {
         if (verbose == 1) {
-          printf("%s.", dns_dntosp(qdn));
+          printf("%s.", dns_dntosp(dn));
           printf(" CNAME %s.\n", dns_dntosp(p.dnsp_dnbuf));
         }
-        qdn = p.dnsp_dnbuf;
+        dns_dntodn(p.dnsp_dnbuf, dn, sizeof(dn));
       }
     }
   }
@@ -507,8 +551,8 @@ static void dnscb(struct dns_ctx *ctx, void *result, void *data) {
   }
   if (verbose < 2) {	/* else it is already printed by dbgfn */
     dns_rewind(&p, NULL);
-    p.dnsp_qtyp = q->qtyp;
-    p.dnsp_qcls = qcls;
+    p.dnsp_qtyp = q->qtyp == DNS_T_ANY ? 0 : q->qtyp;
+    p.dnsp_qcls = qcls == DNS_C_ANY ? 0 : qcls;
     while(dns_nextrr(&p, &rr))
       printrr(&p, &rr);
   }
@@ -527,6 +571,7 @@ int main(int argc, char **argv) {
   struct query *q;
   enum dns_type qtyp = 0;
   struct dns_ctx *nctx = NULL;
+  int flags = 0;
 
   if (!(progname = strrchr(argv[0], '/'))) progname = argv[0];
   else argv[0] = ++progname;
@@ -534,13 +579,13 @@ int main(int argc, char **argv) {
   if (argc <= 1)
     die(0, "try `%s -h' for help", progname);
 
-  if (dns_init(0) < 0 || !(nctx = dns_new(NULL)))
+  if (dns_init(NULL, 0) < 0 || !(nctx = dns_new(NULL)))
     die(errno, "unable to initialize dns library");
   /* we keep two dns contexts: one may be needed to resolve
    * nameservers if given as names, using default options.
    */
 
-  while((i = getopt(argc, argv, "vqt:c:an:o:h")) != EOF) switch(i) {
+  while((i = getopt(argc, argv, "vqt:c:an:o:f:h")) != EOF) switch(i) {
   case 'v': ++verbose; break;
   case 'q': --verbose; break;
   case 't':
@@ -567,9 +612,22 @@ int main(int argc, char **argv) {
     ns[nns++] = optarg;
     break;
   case 'o':
-    if (dns_set_opts(NULL, optarg) != 0)
-      die(0, "invalid option string: `%s'", optarg);
+  case 'f': {
+    char *opt;
+    const char *const delim = " \t,;";
+    for(opt = strtok(optarg, delim); opt != NULL; opt = strtok(NULL, delim)) {
+      if (dns_set_opts(NULL, optarg) == 0)
+        ;
+      else if (strcmp(opt, "aa") == 0) flags |= DNS_AAONLY;
+      else if (strcmp(optarg, "nord") == 0) flags |= DNS_NORD;
+      else if (strcmp(optarg, "dnssec") == 0) flags |= DNS_SET_DO;
+      else if (strcmp(optarg, "do")     == 0) flags |= DNS_SET_DO;
+      else if (strcmp(optarg, "cd") == 0) flags |= DNS_SET_CD;
+      else
+        die(0, "invalid option: `%s'", opt);
+    }
     break;
+  }
   case 'h':
     printf(
 "%s: simple DNS query tool (using udns version %s)\n"
@@ -578,18 +636,22 @@ int main(int argc, char **argv) {
 " -h - print this help and exit\n"
 " -v - be more verbose\n"
 " -q - be less verbose\n"
-" -t type - set query type (A, AA, PTR etc)\n"
+" -t type - set query type (A, AAA, PTR etc)\n"
 " -c class - set query class (IN (default), CH, HS, *)\n"
 " -a - equivalent to -t ANY -v\n"
 " -n ns - use given nameserver(s) instead of default\n"
 "  (may be specified multiple times)\n"
-" -o option:value - set resovler option (the same as setting $RES_OPTIONS):\n"
-"  timeout:sec  - initial query timeout\n"
-"  attempts:num - number of attempt to resovle a query\n"
-"  ndots:num    - if name has more than num dots, lookup it before search\n"
-"  port:num     - port number for queries instead of default 53\n"
-"  udpbuf:num   - size of UDP buffer (use EDNS0 if >512)\n"
-"  (may be specified more than once)\n"
+" -o opt,opt,... (comma- or space-separated list,\n"
+"                 may be specified more than once):\n"
+"  set resovler options (the same as setting $RES_OPTIONS):\n"
+"   timeout:sec  - initial query timeout\n"
+"   attempts:num - number of attempt to resovle a query\n"
+"   ndots:num    - if name has more than num dots, lookup it before search\n"
+"   port:num     - port number for queries instead of default 53\n"
+"   udpbuf:num   - size of UDP buffer (use EDNS0 if >512)\n"
+"  or query flags:\n"
+"   aa,nord,dnssec,do,cd - set query flag (auth-only, no recursion,\n"
+"     enable DNSSEC (DNSSEC Ok), check disabled)\n"
       , progname, dns_version(), progname);
     return 0;
   default:
@@ -613,7 +675,7 @@ int main(int argc, char **argv) {
     sin.sin_port = htons(dns_set_opt(NULL, DNS_OPT_PORT, -1));
     dns_add_serv(NULL, NULL);
     for(i = 0; i < nns; ++i) {
-      if (!inet_aton(ns[i], &sin.sin_addr)) {
+      if (dns_pton(AF_INET, ns[i], &sin.sin_addr) <= 0) {
         struct dns_rr_a4 *rr;
         if (!opened) {
           if (dns_open(nctx) < 0)
@@ -634,7 +696,8 @@ int main(int argc, char **argv) {
       else
         r = dns_add_serv_s(NULL, (struct sockaddr *)&sin);
       if (r < 0)
-        die(errno, "unable to add nameserver %s", inet_ntoa(sin.sin_addr));
+        die(errno, "unable to add nameserver %s",
+             dns_xntop(AF_INET, &sin.sin_addr));
     }
   }
   dns_free(nctx);
@@ -646,6 +709,9 @@ int main(int argc, char **argv) {
   if (verbose > 1)
     dns_set_dbgfn(NULL, dbgcb);
 
+  if (flags)
+    dns_set_opt(NULL, DNS_OPT_FLAGS, flags);
+
   for (i = 0; i < argc; ++i) {
     char *name = argv[i];
     union {
@@ -655,16 +721,18 @@ int main(int argc, char **argv) {
     unsigned char dn[DNS_MAXDN];
     enum dns_type l_qtyp = 0;
     int abs;
-    if (inet_pton(AF_INET, name, &a.addr) > 0) {
+    if (dns_pton(AF_INET, name, &a.addr) > 0) {
       dns_a4todn(&a.addr, 0, dn, sizeof(dn));
       l_qtyp = DNS_T_PTR;
       abs = 1;
     }
-    else if (inet_pton(AF_INET6, name, &a.addr6) > 0) {
+#ifdef HAVE_IPv6
+    else if (dns_pton(AF_INET6, name, &a.addr6) > 0) {
       dns_a6todn(&a.addr6, 0, dn, sizeof(dn));
       l_qtyp = DNS_T_PTR;
       abs = 1;
     }
+#endif
     else if (!dns_ptodn(name, strlen(name), dn, sizeof(dn), &abs))
       die(0, "invalid name `%s'\n", name);
     else
